#!/usr/bin/env python3
"""
File:           gen_data.py
Description:    This file generates data for testing different H-ratios.
License:        This file is licensed under the GNU LGPL V3 license by
                Okke van Eck (2020 - 2023). See the LICENSE file for the
                specifics.
"""

import os
import numpy as np
import random
import math


def new_folder(folder_name):
    """
    Create a new dataset folder and return the path.
    """
    root = "data/"
    dirs = [x[0] for x in os.walk(root)]

    if f"{root}{folder_name}" in dirs:
        raise ValueError(f"Dataset '{folder_name}' already exists.")

    os.mkdir(f"{root}{folder_name}")
    return f"{root}{folder_name}"


def generate_hratio(p_len=25, size=1000):
    """
    Generates the H-ratio dataset given the provided arguments. The provided
    H-ratio range is including h_ratio_begin, and excluding h_ratio_end.
    NOTE: This script has a 1-off bug for lenght 10 H-ratio 1.0!
    :param int      p_len:          Protein length of dataset.
    :param float    h_ratio_begin:  H-ratio fraction range begin (including).
    :param float    h_ratio_end:    H-ratio fraction range end (excluding).
    :param float    h_ratio_step:   H-ratio fraction range step.
    :param int      size:           Number of proteins per H-ratio.
    """
    aminos = ["H", "P"]
    ds_path = "data/vanEck_hratio"

    # Create dataset folder. Throws ValueError if dataset already exists.
    if not os.path.isdir(ds_path):
        os.makedirs(ds_path, exist_ok=True)

    # Construct a list with lower and upper bounds for the h-ratio.
    h_ratio_range = zip(
        np.round(np.arange(0.0, 1.0, 0.1), 2),
        np.round(np.arange(0.1, 1.1, 0.1), 2),
    )

    # Keep track of all produced proteins to prevent duplicates across files.
    all_proteins = set()
    prev_size_all = 0

    for h_ratio_low, h_ratio_high in h_ratio_range:
        cur_fname = f"{''.join(aminos)}{p_len}_r{h_ratio_high}.csv"

        # Only generate dataset if it doesn't exist already.
        if os.path.isfile(f"{ds_path}/{cur_fname}"):
            print(f"Dataset '{cur_fname}' already exists.")
            continue

        # Track number of consecutive iterations no new proteins were produced.
        no_changes = 0

        # Compute maximum of given size and theoretically possible.
        if (
            len(aminos) ** p_len < size
            or math.comb(p_len, math.floor(h_ratio_high * p_len)) < size
        ):
            print(
                f"\n\nCannot produce {size} unique proteins of length {p_len} "
                f"with upper H-ratio of {h_ratio_high}"
            )
            cur_size = math.comb(p_len, math.floor(h_ratio_high * p_len))
            print(f"Producing maximally possible {cur_size} proteins instead.")
        else:
            cur_size = size
            print(
                f"\n\nProducing {cur_size} unique proteins of length {p_len} "
                f"with upper h-ratio of {h_ratio_high}"
            )

        with open(f"{ds_path}/{cur_fname}", "w") as fp:
            cur_set = set()
            fp.write("id,sequence\n")

            # Print debug length every 10 iterations.
            i = 0

            # Generate only unique proteins for this set.
            while len(cur_set) < cur_size:
                new_proteins = set(
                    [
                        "".join(
                            random.choices(
                                aminos,
                                weights=[
                                    min(h_ratio_high, 0.9),
                                    max(round(1 - h_ratio_high, 1), 0.1),
                                ],
                                k=p_len,
                            )
                        )
                        for _ in range(cur_size)
                    ]
                )

                # Filter newly generated proteins on the H-ratio requirements
                # and on previously generated proteins.
                new_proteins = set(
                    [
                        p
                        for p in new_proteins
                        if p.count("P") != 0
                        and h_ratio_low < p.count("H") / p_len <= h_ratio_high
                    ]
                )
                new_proteins -= all_proteins
                all_proteins = all_proteins.union(new_proteins)
                cur_set = cur_set.union(new_proteins)

                # Put cut-off at 1000 iterations of no changes.
                if len(all_proteins) == prev_size_all:
                    no_changes += 1

                    if no_changes == 1000:
                        print("Cutting off due to no changes..")
                        break
                else:
                    prev_size_all = len(all_proteins)
                    no_changes = 0

                # Print debug if needed, update tracker.
                if i % 250 == 0:
                    print(f"    {h_ratio_high} - cur_size:  {len(cur_set)}")
                    print("\tFilter:   " f"({h_ratio_low}, {h_ratio_high}]")
                    print(
                        f"\tWeights:  [{min(h_ratio_high, 0.9)}, "
                        f"{max(round(1 - h_ratio_high, 1), 0.1)}]"
                    )
                i += 1

            # Write newly generated set to file.
            for i, p in enumerate(cur_set):
                fp.write(f"{i},{p}\n")


if __name__ == "__main__":
    # Fetch arguments from user required for generating data.
    print(
        "Requesting needed information.\nLeave blank for the default " "value."
    )
    print("==================================", end="\n\n")

    p_len_str = input("Protein length (default=25): ").strip()
    p_len = int(p_len_str) if p_len_str else 25

    size_str = input("Maximum #proteins per set (default=1000): ").strip()
    size = int(size_str) if size_str else 1000

    # Generate the data using the provided arguments.
    generate_hratio(p_len, size)



#!/usr/bin/env python3
"""
File:           visualize.py
Description:    This file contains functions for visualizing a Protein
                instance.
License:        This file is licensed under the GNU LGPL V3 license by
                Okke van Eck (2020 - 2023). See the LICENSE file for the
                specifics.
"""

from .helpers import get_ordered_positions, get_scoring_pairs

from matplotlib import pyplot as plt
from matplotlib.ticker import MaxNLocator
from matplotlib.lines import Line2D
import seaborn as sns
import pandas as pd


def _plot_aminos_2d_basic(protein, df, ax):
    """
    Plot amino acids in basic style in a 2D figure.
    :param Protein      protein:    Protein object to plot the hash of.
    :param DataFrame    df:         DataFrame with all ordered positions.
    :param Axes         ax:         Axis to plot on.
    """
    ax.plot(df["x"], df["y"], color="black", alpha=0.65, zorder=1)
    sns.scatterplot(
        x="x",
        y="y",
        data=df,
        hue="Type",
        hue_order=["H", "P"],
        style="Type",
        markers={"H": "o", "P": "s"},
        palette={"H": "royalblue", "P": "orange"},
        s=80,
        zorder=2,
        ax=ax,
    )

    # Plot dotted lines between the aminos that increase the stability.
    pairs = get_scoring_pairs(protein)

    for pos1, pos2 in pairs:
        ax.plot(
            [pos1[0], pos2[0]],
            [pos1[1], pos2[1]],
            linestyle=":",
            color="indianred",
            alpha=0.9,
            zorder=1,
            lw=2,
        )


def _plot_aminos_2d_paper(
    protein, df, ax, linewidth, markersize, annotate_first=False
):
    """
    Plot amino acids in paper style in a 2D figure.
    :param Protein      protein:    Protein object to plot the hash of.
    :param DataFrame    df:         DataFrame with all ordered positions.
    :param Axes         ax:         Axis to plot on.
    :param float        linewidth:      Width of the lines.
    :param float        markersize:     Size of the markers.
    """
    # Split dataframe on amino acid type.
    df_H = df.loc[df["Type"] == "H"]
    df_P = df.loc[df["Type"] == "P"]

    ax.plot(
        df["x"],
        df["y"],
        color="black",
        alpha=0.65,
        linewidth=linewidth,
        zorder=1,
    )
    sns.scatterplot(
        x="x",
        y="y",
        data=df_H,
        marker="o",
        edgecolor="royalblue",
        s=markersize,
        zorder=2,
        ax=ax,
        label="H",
    )
    sns.scatterplot(
        x="x",
        y="y",
        data=df_P,
        marker="o",
        facecolor="white",
        edgecolor="orange",
        linewidth=2.5,
        s=markersize,
        zorder=2,
        ax=ax,
        label="P",
    )

    # Plot first point with a green edge color.
    if annotate_first:
        # Plot first point with a different color.
        if df.iloc[0]["Type"] == "H":
            ax.scatter(
                df.iloc[0]["x"],
                df.iloc[0]["y"],
                marker="o",
                fc="royalblue",
                ec="#00ce00",
                lw=2.5,
                s=markersize,
                zorder=2,
            )
        else:
            ax.scatter(
                df.iloc[0]["x"],
                df.iloc[0]["y"],
                marker="o",
                fc="white",
                ec="#00ce00",
                lw=2.5,
                s=markersize,
                zorder=2,
            )

    # Plot dotted lines between the aminos that increase the stability.
    pairs = get_scoring_pairs(protein)

    for pos1, pos2 in pairs:
        ax.plot(
            [pos1[0], pos2[0]],
            [pos1[1], pos2[1]],
            linestyle=":",
            color="indianred",
            alpha=0.9,
            linewidth=linewidth,
            zorder=1,
        )

    # Remove axis, and position legend in the upper right with created space.
    ax.axis("off")


def _plot_aminos_3d_basic(protein, df, ax):
    """
    Plot amino acids in basic style in a 3D figure.
    :param Protein      protein:    Protein object to plot the hash of.
    :param DataFrame    df:         DataFrame with all ordered positions.
    :param Axes         ax:         Axis to plot on.
    """
    # Split dataframe on amino acid type.
    df_H = df.loc[df["Type"] == "H"]
    df_P = df.loc[df["Type"] == "P"]

    ax.plot(df["x"], df["y"], df["z"], color="black", alpha=0.65, zorder=1)

    # Plot the aminos connected with an opaque line.
    ax.scatter(
        df_H["x"],
        df_H["y"],
        df_H["z"],
        c="royalblue",
        marker="o",
        depthshade=False,
        s=60,
        label="H",
    )
    ax.scatter(
        df_P["x"],
        df_P["y"],
        df_P["z"],
        c="orange",
        marker="s",
        depthshade=False,
        s=60,
        label="P",
    )

    # Plot dotted lines between the aminos that increase the stability.
    pairs = get_scoring_pairs(protein)

    for pos1, pos2 in pairs:
        ax.plot(
            [pos1[0], pos2[0]],
            [pos1[1], pos2[1]],
            [pos1[2], pos2[2]],
            linestyle=":",
            color="indianred",
            alpha=0.9,
            zorder=1,
            lw=2,
        )


def _plot_aminos_3d_paper(
    protein, df, ax, linewidth, markersize, annotate_first=False
):
    """
    Plot amino acids in paper style in a 3D figure.
    :param Protein      protein:    Protein object to plot the hash of.
    :param DataFrame    df:         DataFrame with all ordered positions.
    :param Axes         ax:         Axis to plot on.
    """
    # Split dataframe on amino acid type.
    df_H = df.loc[df["Type"] == "H"]
    df_P = df.loc[df["Type"] == "P"]

    ax.plot(
        df["x"],
        df["y"],
        df["z"],
        color="black",
        alpha=0.65,
        linewidth=linewidth,
        zorder=1,
    )

    ax.scatter(
        df_H["x"],
        df_H["y"],
        df_H["z"],
        marker="o",
        edgecolor="royalblue",
        s=markersize,
        zorder=2,
        label="H",
        depthshade=False,
    )
    ax.scatter(
        df_P["x"],
        df_P["y"],
        df_P["z"],
        marker="o",
        facecolor="white",
        edgecolor="orange",
        lw=1.3,
        s=markersize,
        zorder=2,
        label="P",
        depthshade=False,
    )

    # Plot first point with a green edge color.
    if annotate_first:
        if df.iloc[0]["Type"] == "H":
            ax.scatter(
                df.iloc[0]["x"],
                df.iloc[0]["y"],
                df.iloc[0]["z"],
                marker="o",
                fc="royalblue",
                ec="#00ce00",
                lw=1.3,
                s=markersize,
                zorder=2,
            )
        else:
            ax.scatter(
                df.iloc[0]["x"],
                df.iloc[0]["y"],
                df.iloc[0]["z"],
                marker="o",
                fc="white",
                ec="#00ce00",
                lw=1.3,
                s=markersize,
                zorder=2,
            )

    # Plot dotted lines between the aminos that increase the stability.
    pairs = get_scoring_pairs(protein)

    for pos1, pos2 in pairs:
        ax.plot(
            [pos1[0], pos2[0]],
            [pos1[1], pos2[1]],
            [pos1[2], pos2[2]],
            linestyle=":",
            color="indianred",
            alpha=0.9,
            zorder=1,
            lw=linewidth,
        )

    # Remove axis, and position legend in the upper right with created space.
    ax.axis("off")


def plot_protein(
    protein,
    style="basic",
    ax=None,
    legend=True,
    legend_style="inner",
    show=True,
    linewidth=2.5,
    markersize=210,
    annotate_first=False,
):
    """
    Plot conformation of a protein.
    :param Protein  protein:        Protein object to plot the hash of.
    :param [str]    style:          What style to plot the proteins in.
    :param Axes     ax:             Axis to plot Protein on.
    :param bool     legend:         True if a legend needs to be added.
    :param str      legend_style:   Either 'inner' or 'outer'.
    :param bool     show:           True if plot.show() needs to be called.
    :param float    linewidth:      Width of the lines.
    :param float    markersize:     Size of the markers.
    :param bool     annotate_first: True if first amino acids needs annotation.
    """
    # Catch unplottable dimensions.
    if protein.dim != 2 and protein.dim != 3:
        raise RuntimeError(
            f"Cannot plot the structure of a protein with "
            f"dimension '{protein.dim}'"
        )

    # Create axis to plot onto if not given.
    if ax is None:
        if style == "paper":
            fig = plt.figure(figsize=(5, 3.125))
        else:
            fig = plt.figure(figsize=(5, 6))
            sns.set_style("whitegrid")

        if protein.dim == 2:
            ax = fig.gca()
        else:
            ax = fig.add_subplot(projection="3d")

    # Fetch data in right dimension.
    if protein.dim == 2:
        df = pd.DataFrame(
            get_ordered_positions(protein), columns=["x", "y", "Type"]
        )
        df = df.astype({"x": "int32", "y": "int32"})
    else:
        df = pd.DataFrame(
            get_ordered_positions(protein), columns=["x", "y", "z", "Type"]
        )
        df = df.astype({"x": "int32", "y": "int32", "z": "int32"})

    # Plot the selected style.
    if style == "paper":
        if protein.dim == 2:
            _plot_aminos_2d_paper(
                protein, df, ax, linewidth, markersize, annotate_first
            )
        else:
            _plot_aminos_3d_paper(
                protein, df, ax, linewidth, markersize, annotate_first
            )
    elif style == "basic":
        ax.set_xlabel("x-axis", fontsize=13)
        ax.set_ylabel("y-axis", fontsize=13)
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))

        # Plot dimension specific.
        if protein.dim == 2:
            ax.set_title(f"2D conformation with {protein.score} energy")
            _plot_aminos_2d_basic(protein, df, ax)
        else:
            ax.set_title(f"3D conformation with {protein.score} energy")
            ax.set_zlabel("z-axis", fontsize=13)
            ax.zaxis.set_major_locator(MaxNLocator(integer=True))
            _plot_aminos_3d_basic(protein, df, ax)

    # If adding legend, remove title from legend and add item for bonds.
    if legend:
        handles, labels = ax.get_legend_handles_labels()

        # Only add bond label if actually in plot.
        if protein.score != 0:
            score_patch = Line2D(
                [],
                [],
                color="indianred",
                linestyle=":",
                linewidth=linewidth,
                alpha=0.9,
                label="Bond",
            )
            handles.append(score_patch)
            labels.append(score_patch.get_label())

        # Style legend according to plotting style.
        if (
            style == "paper" and legend_style == "outer"
        ) or legend_style == "outer":
            box = ax.get_position()
            ax.set_position([box.x0, box.y0, box.width * 0.7, box.height])
            ax.legend(
                handles=handles,
                labels=labels,
                loc="upper left",
                bbox_to_anchor=(1, 1),
            )
        else:
            ax.legend(handles=handles, labels=labels, prop={"size": 12})
    elif protein.dim == 2:
        ax.get_legend().remove()

    # Show plot if specified.
    if show:
        plt.show()



#!/usr/bin/env python3
"""
File:           datasets.py
Description:    This file contains functions for loading datasets.
License:        This file is licensed under the GNU LGPL V3 license by
                Okke van Eck (2020 - 2023). See the LICENSE file for the
                specifics.
"""

import pandas as pd


def _load_dataset(folder, filename):
    """Returns a specified dataset as a dataframe."""
    return pd.read_csv(f"prospr/data/{folder}/{filename}")


def load_vanEck250(length=10):
    """Returns a vanEck250 dataset as a pandas dataframe."""
    return _load_dataset("vanEck250", f"HP{length}.csv")


def load_vanEck1000(length=10):
    """Returns a vanEck1000 dataset as a pandas dataframe."""
    return _load_dataset("vanEck1000", f"HP{length}.csv")


def load_vanEck_hratio(length=25, hratio=0.1):
    """Returns a vanEck_hratio dataset as a pandas dataframe."""
    return _load_dataset("vanEck_hratio", f"HP{length}_r{hratio}.csv")



#!/usr/bin/env python3
"""
File:           helpers.py
Description:    This file contains helper functions for a Protein instance.
License:        This file is licensed under the GNU LGPL V3 license by
                Okke van Eck (2020 - 2023). See the LICENSE file for the
                specifics.
"""

import numpy as np


def get_scoring_aminos(protein):
    """
    Collect all placed aminos from a protein that may score points.
    """
    score_pos = {}
    cur_pos = [0 for _ in range(protein.dim)]
    amino_acid = protein.get_amino(cur_pos)
    idx = amino_acid.index
    next_dir = amino_acid.next_move
    max_weights = protein.max_weights

    # Store origin if it may score points.
    if max_weights[idx] < 0:
        score_pos[tuple(cur_pos)] = np.array([0, next_dir], dtype=np.int64)

    while next_dir != 0:
        # Compute position of next amino and get its amino and fold.
        cur_pos[abs(next_dir) - 1] += next_dir // abs(next_dir)
        amino_acid = protein.get_amino(cur_pos)
        idx = amino_acid.index
        fold = amino_acid.next_move

        # Store previous directions for checking existing connections.
        prev_dir = -next_dir
        next_dir = fold

        # Save amino if it may score points.
        if max_weights[idx] < 0:
            score_pos[tuple(cur_pos)] = np.array(
                [prev_dir, next_dir], dtype=np.int64
            )

    return score_pos


def get_scoring_pairs(protein):
    """
    Fetch positions of scoring amino in pairs for the currently folded amino
    acid chain.
    """
    # Get dictionary with the amino's that can possibly score points.
    score_aminos = get_scoring_aminos(protein)

    # Sort positions from bottom-left to upper-right.
    moves = np.array([m for m in range(1, protein.dim + 1)])
    pairs = np.empty((1, 2, protein.dim), dtype=np.int64)

    # Check if a score can be made with the amino above or to the right.
    for pos, [prev_dir, next_dir] in score_aminos.items():
        other_pos = np.array(pos, dtype=np.int64)

        # Check every positive direction, started in the lower quadrant.
        for move in moves:
            if prev_dir != move and next_dir != move:
                other_pos[move - 1] += 1

                if tuple(other_pos) in score_aminos:
                    new_pair = np.array(
                        [[pos, tuple(other_pos)]], dtype=np.int64
                    )
                    pairs = np.vstack((pairs, new_pair))

                other_pos[move - 1] -= 1

    # Return all pairs without the first row of zeros.
    return pairs[1:]


def get_ordered_positions(protein):
    """
    Fetch the positions of the currently folded amino chain in order of
    placement.
    """
    # Fetch done moves and the aminos used so far.
    moves = protein.hash_fold()
    aminos = protein.sequence[: len(moves) + 1]

    # Setup storage of the amino positions.
    cur_pos = np.array([0 for _ in range(protein.dim)], dtype=np.int64)
    positions = np.array([*cur_pos, aminos[0]])

    # Save the coordinates of the aminos in order.
    for amino, move in zip(aminos[1:], moves):
        cur_pos[abs(move) - 1] += move // abs(move)
        positions = np.vstack((positions, np.array([*cur_pos, amino])))

    return positions



from prospr_core import (
    AminoAcid,
    Protein,
    depth_first,
    depth_first_bnb,
    beam_search,
)
from .datasets import load_vanEck250, load_vanEck1000, load_vanEck_hratio
from .visualize import plot_protein

# Import __version__ from _version.py during compile time.
exec(open("prospr/_version.py").read())

__all__ = [
    "__version__",
    "AminoAcid",
    "Protein",
    "depth_first",
    "depth_first_bnb",
    "beam_search",
    "load_vanEck250",
    "load_vanEck1000",
    "load_vanEck_hratio",
    "plot_protein",
]



#!/usr/bin/env python3
"""
File:           setup.py
Description:    This file contains the setup required for distributing the
                package on PyPi.org.
License:        This file is licensed under the GNU LGPL V3 license by
                Okke van Eck (2020 - 2023). See the LICENSE file for the
                specifics.
"""

from setuptools import setup
from pybind11.setup_helpers import Pybind11Extension, build_ext


setup(
    ext_modules=[
        Pybind11Extension(
            name="prospr_core",
            sources=["prospr/core/core_module.cpp"],
            language="c++",
        ),
    ],
    cmdclass={"build_ext": build_ext},
)
